/*
 * lib/arc.c
 * ARC BIOS interface routines
 *
 * Copyright (C) 1994, 1995 Waldorf Electronics,
 * written by Andreas Busse
 */
#include <stand/debug.h>
#include <stand/stdio.h>
#include <arc/types.h>
#include <arc/pathname.h>
#include <arc/proto.h>
#include <arc/stdlib.h>
#include <arc/arc.h>

#include <asm/bootinfo.h>

/*
 * Prototypes
 */
void print_component(CONFIGURATION_COMPONENT *Component);
CONFIGURATION_COMPONENT *find_component(CONFIGURATION_COMPONENT *top,
					CONFIGURATION_CLASS class,
					CONFIGURATION_TYPE type);

/*
 * Global variables
 */
char noinfo[] = "No information available";

/*
 * getVendorId()
 * Get vendor identification from ARC Bios 
 */
char *getVendorId(void)
{
  SYSTEMID *SystemId;		/* system info from ARCS BIOS */

  SystemId = (SYSTEMID *)GetSystemId();
  return((char *)SystemId->VendorId);
}

/*
 * getProductId()
 * Get product identification from ARC Bios 
 */
char *getProductId(void)
{
  SYSTEMID *SystemId;		/* system info from ARCS BIOS */

  SystemId = (SYSTEMID *)GetSystemId();
  return((char *)SystemId->ProductId);
}

/*
 * getSystemId()
 * Get System Identification from ARC BIOS
 */
char *getSystemId(void)
{
  CONFIGURATION_COMPONENT *component;

  component = GetChild(NULL);	/* get top element from config tree */
  if (!component)
    return(noinfo);
  else
    return((char *)component->Identifier);
}

/*
 * getCpuType()
 * Get CPU Type from ARC BIOS
 */
char *getCpuType(void)
{
  CONFIGURATION_COMPONENT *component;

  component = GetChild(NULL);	/* get top element from config tree */
  component = find_component(component,ProcessorClass,CentralProcessor);
  if (!component)
    return((char *)&noinfo);
  else
    return((char *)component->Identifier);
}

#if 0
/*
 * getCpuId()
 * Get CPU ID from ARC BIOS
 */
unsigned int getCpuId(void)
{
  CONFIGURATION_COMPONENT *component;

  component = GetChild(NULL);	/* get top element from config tree */
  component = find_component(component,ProcessorClass,CentralProcessor);
  if (!component)
    return(component->Key);
  else
    return(CPU_UNKNOWN);
}

/*
 * getFpuType()
 * Get FPU Type from ARC BIOS
 */
char *getFpuType(void)
{
  CONFIGURATION_COMPONENT *component;

  component = GetChild(NULL);	/* get top element from config tree */
  component = find_component(component,ProcessorClass,FloatingPointProcessor);
  if (!component)
    return((char *)&noinfo);
  else
    return((char *)component->Identifier);
}

/*
 * getPrimaryICacheSize()
 * Get Primary Instruction Cache Size from ARC BIOS
 */
unsigned int getPrimaryICacheSize(void)
{
  CONFIGURATION_COMPONENT *component;
  int refill_size;
  int line_size;
  int cache_size;

  component = GetChild(NULL);	/* get top element from config tree */
  component = find_component(component,CacheClass,PrimaryIcache);
  if (component)
    {
      refill_size = 1 << (component->Key >> 24);
      line_size = 1 << ((component->Key >> 16) && 0xff);
      cache_size = 0x1000 << (component->Key & 0xffff);
      return(cache_size);
    }
  return(SIZE_UNKNOWN);
}

/*
 * getPrimaryDCacheSize()
 * Get Primary Data Cache Size from ARC BIOS
 */
unsigned int getPrimaryDCacheSize(void)
{
  CONFIGURATION_COMPONENT *component;
  int refill_size;
  int line_size;
  int cache_size;
  
  component = GetChild(NULL);	/* get top element from config tree */
  component = find_component(component,CacheClass,PrimaryDcache);
  if (component)
    {
      refill_size = 1 << (component->Key >> 24);
      line_size = 1 << ((component->Key >> 16) && 0xff);
      cache_size = 0x1000 << (component->Key & 0xffff);
      return(cache_size);
    }
  return(SIZE_UNKNOWN);
}
#endif

/*
 * getSecondaryICacheSize()
 * Get Secondary Instruction Cache Size from ARC BIOS
 */
unsigned int getSecondaryICacheSize(void)
{
  CONFIGURATION_COMPONENT *component;
  int size;
  
  component = GetChild(NULL);	/* get top element from config tree */
  component = find_component(component,CacheClass,SecondaryIcache);
  if (component)
    {
      size = 0x1000 << (component->Key & 0xffff);
      return(size);
    }
  return(SIZE_UNKNOWN);
}

/*
 * getSecondaryDCacheSize()
 * Get Secondary Data Cache Size from ARC BIOS
 */
unsigned int getSecondaryDCacheSize(void)
{
  CONFIGURATION_COMPONENT *component;
  int size;
  
  component = GetChild(NULL);	/* get top element from config tree */
  component = find_component(component,CacheClass,SecondaryDcache);
  if (component)
    {
      size = 0x1000 << (component->Key & 0xffff);
      return(size);
    }
  return(SIZE_UNKNOWN);
}

/*
 * getSecondaryCacheSize()
 * Get Shared Secondary Cache Size from ARC BIOS
 */
unsigned int getSecondaryCacheSize(void)
{
  CONFIGURATION_COMPONENT *component;
  unsigned int refill_size;
  unsigned int line_size;
  unsigned int cache_size;
  
  component = GetChild(NULL);	/* get top element from config tree */
  component = find_component(component,CacheClass,SecondaryCache);
  if (component)
    {
      refill_size = 1 << (component->Key >> 24);
      line_size = 1 << ((component->Key >> 16) && 0xff);
      cache_size = 0x1000 << (component->Key & 0xffff);
      return(cache_size);
    }
  return(SIZE_UNKNOWN);
}

/*
 * getDisplayInfo()
 * Get Video Display Info from ARC BIOS
 */
DisplayInfo *getDisplayInfo()
{
    P_DISPLAY_STATUS ds;
    static DisplayInfo di;

    if (debuglevel >= 2)
    {
	printf("getDisplayInfo(): calling ARC GetDisplayStatus()\r\n");
	fflush(stdout);
    }

    /*
     * GetDisplayStatus() seems to be broken in
     * the Mips Magnum 4000 BIOS, so we fill in
     * some guesses by hand.
     * FIXME: only true for 1024x768 mode !
     */
    if (__hw_type == MACH_MIPS_MAGNUM_4000)
    {
	di.cursor_x = 0;		/* start at col. 0, line 40 */
	di.cursor_y = 40;
	di.columns  = 128;	        /* 128 * 8 = 1024 pixels */
	di.lines    = 48;		/* 48 * 16 =  768 pixels */
	if (debuglevel >= 2)
	{
	    printf("DisplayStatus: x=%d, y=%d, columns=%d, lines=%d\r\n",
		   di.cursor_x,
		   di.cursor_y,
		   di.columns,
		   di.lines);
	    fflush(stdout);
	}
	return &di;
    }
    
    if((ds = GetDisplayStatus(stdout->fh)))
    {
	di.cursor_x = ds->CursorXPosition - 1;
	di.cursor_y = ds->CursorYPosition - 1;
	di.columns  = ds->CursorMaxXPosition;
	di.lines    = ds->CursorMaxYPosition;
	if (debuglevel >= 2)
	{
	    printf("DisplayStatus: x=%d, y=%d, columns=%d, lines=%d\r\n",
		   di.cursor_x,
		   di.cursor_y,
		   di.columns,
		   di.lines);
	    fflush(stdout);
	}
	return &di;
    }

    return NULL;
}
 
/*
 * find_component()
 * Find configuration info about a certain component
 */
CONFIGURATION_COMPONENT *find_component(
    CONFIGURATION_COMPONENT *top,
    CONFIGURATION_CLASS class,
    CONFIGURATION_TYPE type)
{
    CONFIGURATION_COMPONENT *cc;
    static int calls = 0;
  
    cc = top;
    calls++;
    while(cc)
    {
/*      printf("[%d] Current Component:\r\n",calls); */
/*      print_component(cc); */
      if (cc->Class == class && cc->Type == type)
	{
/*	  printf("[%d] Matching component (class=%d, type=%d) found.\r\n",
		 calls,class,type); */
	  break;
	}
/*      printf("[%d] Not matching. Press a key for next...\r\n",calls); */
/*      getchar(); */
      cc = find_component(GetChild(top),class,type); /* step down */
      if (cc != NULL)
	break;
      cc = top = GetPeer(top);	/* no, get next peer */
    }
  calls--;
  return(cc);
}
  
/*
 * print_component()
 * Print Component Info (used only for debugging purposes)
 */
void print_component(CONFIGURATION_COMPONENT *component)
{
    if (!component)
        return;
  
    printf("Class=%d, Type=%d, Flags=%d, Version=%d, Key=%d, Mask=%d, Size=%d, Length=%d\r\n",
           component->Class,
           component->Type,
           component->Flags,
           component->Version,
           component->Key,
           component->AffinityMask,
           component->ConfigurationDataLength,
           component->IdentifierLength);

    printf("Identifier = %s\r\n",component->Identifier);
}
